/* ReSCAL - Global macro definitions
 *
 * Copyright (C) 2011
 *
 * Author: Olivier Rozier <rozier@ipgp.fr>
 *
 * This file is part of ReSCAL.
 *
 * ReSCAL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#ifdef _MAIN_
long total_memory=0;  //size totale de memoire allouee
#ifdef LOG_FILE
FILE *log_file=NULL;
#endif
#else
extern long total_memory;
#ifdef LOG_FILE
extern FILE *log_file;
#endif
#endif //_MAIN_


#ifdef LOG_FILE
#define LogPrintf(...) { \
    if (log_file) \
      {fprintf(log_file, __VA_ARGS__); fflush(log_file);} \
    else \
      {printf(__VA_ARGS__); fflush(stdout);} \
    }
#else
#define LogPrintf(...) {printf(__VA_ARGS__); fflush(stdout);}
//#define Log(...) fprintf(stderr, __VA_ARGS__)
#endif //LOG_FILE

#define WarnPrintf(...) { \
    LogPrintf(__VA_ARGS__); \
    fprintf(stderr, __VA_ARGS__); \
    fflush(stderr); \
  }

#define ErrPrintf(...) { \
    LogPrintf(__VA_ARGS__); \
    LogPrintf("%s (%d)\n", __FILE__, __LINE__); \
    fprintf(stderr, __VA_ARGS__); \
    fprintf(stderr, "%s (%d)\n", __FILE__, __LINE__); \
    fflush(stderr); \
  }

#define Assert(_COND) { \
    if(!(_COND)){ \
      ErrPrintf("ERROR: assertion failed in "); \
      assert(_COND); \
    } \
}


#define AllocMemory(addr,type,size) \
{			\
  if (size < 0) {     \
    ErrPrintf("ERROR: bad argument %ld in malloc function\n",(long)size);		\
    exit(EXIT_FAILURE);                           \
  }						\
  if( (addr=(type *)malloc(sizeof(type)*size)) == NULL ){	\
    ErrPrintf("ERROR: failed to allocate %lu bytes\n",(unsigned long)(sizeof(type)*size));		\
    exit(EXIT_FAILURE);                           \
  }						\
  total_memory += sizeof(type)*size;          \
}

#define AllocAlignedMemory(addr,type,size,align) \
{			\
  if (size < 0) {     \
    ErrPrintf("ERROR: bad argument %ld in malloc function\n",(long)size);		\
    exit(EXIT_FAILURE);                           \
  }						\
  if (posix_memalign( (void**)&addr, align, sizeof(type)*size)){	\
    ErrPrintf("ERROR: failed to allocate and align %lu bytes\n",(unsigned long)(sizeof(type)*size));		\
    exit(EXIT_FAILURE);                           \
  }						\
  total_memory += sizeof(type)*size;          \
}

#define ReallocMemory(addr,type,size) \
{			\
  if (size < 0) {     \
    ErrPrintf("ERROR: bad argument %ld in realloc function\n",(long)size);		\
    exit(EXIT_FAILURE);                           \
  }						\
  if( (addr=(type *)realloc(addr,sizeof(type)*size)) == NULL ){	\
    ErrPrintf("ERROR: failed to reallocate %lu bytes\n",(unsigned long)(sizeof(type)*size));		\
    exit(EXIT_FAILURE);                           \
  }						\
}

#define ResetMemory(addr,type,size) memset(addr,0,sizeof(type)*size);

#define PrintMemory(var,type,size,flag) \
  long size_mem = sizeof(type)*size; \
  char *_str = flag ? "re" : ""; \
  if (size_mem >= 1000000) \
    LogPrintf("%sallocated memory for %s: %.2f Mo\n", _str, var, (float)size_mem/1000000) \
  else \
    LogPrintf("%sallocated memory for %s: %ld\n", _str, var, size_mem)

#define AllocMemoryPrint(var,addr,type,size) \
{ \
  PrintMemory(var,type,size,0); \
  AllocMemory(addr,type,size); \
}

#define AllocAlignedMemoryPrint(var,addr,type,size,align) \
{ \
  PrintMemory(var,type,size,0); \
  AllocAlignedMemory(addr,type,size,align); \
}

#define ReallocMemoryPrint(var,addr,type,size) \
{ \
  PrintMemory(var,type,size,1); \
  ReallocMemory(addr,type,size); \
}

#define PrintTotalMemory() \
{ \
  if (total_memory >= 1000000000) \
    LogPrintf("total allocated memory: %.2f Go\n", (float)total_memory/1000000000) \
  else if (total_memory >= 1000000) \
    LogPrintf("total allocated memory: %.2f Mo\n", (float)total_memory/1000000) \
  else \
    LogPrintf("total allocated memory: %ld\n", total_memory) \
}

#define FreeMemory(addr,type,size) \
{ \
  if (addr == NULL){ \
    ErrPrintf("Error in FreeMemory\n");	\
    exit(EXIT_FAILURE);                           \
  }						\
  free(addr); \
  total_memory -= sizeof(type)*size;          \
}

#define Round(x) ((x)>=0 ? (int)((x)+0.5) : (int)((x)-0.5))

#define Calcule_xyz(index, x, y, z) \
{                           \
  int res;                  \
  z = (int)index/HL;        \
  res = index - z*HL;       \
  y = (int)res/L;           \
  x = res - y*L;            \
}

#define Rotate_xz(x, z, angle) \
{                               \
  float alpha = angle*PI/180.0; \
  float co = cos(alpha);        \
  float si = sin(alpha);        \
  float dx0 = L/2.0 - 0.5;        \
  float dz0 = D/2.0 - 0.5;        \
  float dx = x - dx0;           \
  float dz = z - dz0;           \
  x = Round(dx0 + co*dx - si*dz); \
  z = Round(dz0 + si*dx + co*dz); \
}

#define Align4(n) n = (n+3) & 0xfffffffc

#define Check_min(_Val,_Min) _Val = ((_Val<_Min)? _Min :  _Val)

#define Check_max(_Val,_Max) _Val = ((_Val>_Max)? _Max :  _Val)

#define Check_min_max(_Val,_Min,_Max) _Val = ((_Val<_Min)? _Min : (_Val>_Max)? _Max : _Val)

#define Min(_A,_B) ((_A<=_B)? _A : _B)

#define Max(_A,_B) ((_A>=_B)? _A : _B)

#define sgn(x) ((x<0)?-1:((x>0)?1:0))

#define Modulo(x,y) \
{ \
  Assert(y>0); \
  while(x>=y) x-=y; \
  while(x<0) x+=y; \
}

/// Number of decimal digits
#define Nb_digits(_x, _nb) \
{                   \
    long _n=9;       \
    for(_nb=1; _n<_x; _nb++) _n = 10*_n+9;   \
}

#define CheckAlign32(_addr) (((long)(_addr) & 0x1f)==0)

#ifdef GTK1
#define g_timeout_add_seconds(interval, function, data) \
  gtk_timeout_add(interval*1000, function, data)

#define g_timeout_add_seconds_full(pr, interval, function, data, notify) \
  gtk_timeout_add_full(pr, interval*1000, function, data, notify)
#endif

/*
#ifdef __CYGWIN__
# include <limits.h>
# define drand48() ((double)rand()/INT_MAX)
# define srand48(sv) (srand((unsigned)(sv)))
#endif
*/
